/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 2004 James Bunton <james@delx.cjb.net>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *	
 * -------------------------------------------------------------------------- */

#ifndef FILETRANSFER_H
#define FILETRANSFER_H

#include "msntrans.h"
#include "session.h"
#include "sb.h"

#define MSNFTP_GUID "{5D3E02AB-6190-11d3-BBBB-00C04F795683}"

/* Order of stages for an incoming file, where the user does have Connectivity
INCOMING_INVITE_RECEIVED		- Received the initial invitation packet with invite-command: INVITE
INCOMING_ACCEPT_SENT			- Sent out an accept packet, waiting for them to offer to be a server
INCOMING_TRANSFERSTARTED	- They sent another invitation packet with an offer to be server, we connect to their IP
INCOMING_SENTVER
INCOMING_SENTUSR
INCOMING_BINARYMODE
TRANSFERDONE

INCOMING_INVITE_RECEIVED		- Received the initial invitation packet with invite-command: INVITE and connectivity: N
INCOMING_WAITING_FOR_IP		- We sent an ACCEPT packet requesting their IP address (so that we can identify them when they connect)
INCOMING_OFFERTOSERVE_SENT	- We received their IP and offered to serve
INCOMING_TRANSFERSTARTED	- They sent another invitation packet with an offer to be server, we connect to their IP
INCOMING_SENTVER
INCOMING_SENTUSR
INCOMING_BINARYMODE
TRANSFERDONE
*/

typedef enum {
	NOTHING,
	
	INCOMING_INVITE_RECEIVED,
	INCOMING_ACCEPT_SENT,
	INCOMING_WAITING_FOR_IP,
	INCOMING_OFFERTOSERVE_SENT,
	INCOMING_TRANSFERSTARTED,
	INCOMING_SENTVER,
	INCOMING_SENTUSR,
	INCOMING_BINARYMODE,
	
	OUTGOING_INVITE_SENT,
	OUTGOING_ACCEPT_RECEIVED,
	OUTGOING_OFFERTOSERVE_SENT,
	OUTGOING_TRANSFERSTARTED,
	OUTGOING_BINARYMODE,
	
	TRANSFERDONE
} msnftp_stage;

/* A typical order of stages for receiving a file is
INCOMING_INVITE
OUTGOING_ACCEPT
INCOMING_ACCEPT
*/

typedef struct msnftp_st
{
	session s;
	pool p;
	char *remoteperson;
	char *cancelcode; // If we ever need to cancel
	
	// Our private switchboard
	sbchat privatesb;
	spool sbbuff; // switchboard buffer, while we wait for privatesb to be ready
	
	// Out of bound file receiving stream
	mio oob_m;
	
	// Needed at each stage of the invitation
	msnftp_stage stage;
	char *invitationcookie;
	
	// Needed at the INVITE (first) stage
	char *applicationguid; // Identifies us as a file transfer
	char *applicationfile; // The file name
	char *applicationfilesize;
	int connectivity;
	char *remoteipaddress;
	
	// Needed at ACCEPT stage
	char *ipaddress; // The IP address
	char *ipaddressalt; // Secondary IP address
	char *port;
	char *portalt; // Secondary port
	char *authcookie; // Unique cookie used in the actual transfer
	
	// Needed during the transfer stage
	int totalbytes;
	int bytesreceived;
	int bytestonext; // Number of bytes until the next header
	int headergrabbed;
	char *filedata;
	
} *msnftp, _msnftp;


/* File receiving */
msnftp mt_msnftp_receive_create(session s, char *remoteperson, char *text, sbchat sc);
void mt_msnftp_receive_accept(msnftp ft);
void mt_msnftp_receive_starttransfer(msnftp ft, char *body);
void mt_msnftp_receive_offertoserve(msnftp ft, char *body);
void mt_msnftp_receive_finished(msnftp ft);

/* Switchboard negotiation */
void mt_msnftp_sb_handle(session s, char *remoteperson, char *body, sbchat sc);
void mt_msnftp_sb_checkprivate(msnftp ft);
void mt_msnftp_sb_sendmessage(msnftp ft, char *text);
void mt_msnftp_sb_flush(session s, char *invitationcookie);

/* Out of band data functions */
void mt_msnftp_oob_connect(msnftp ft);
void mt_msnftp_oob_send(msnftp ft, char *data);
void mt_msnftp_oob_in(mio m, int state, void *arg, char *buffer, int bufsz);
void mt_msnftp_oob_listen(mio m, int state, void *arg, char *buffer, int bufsz);


#endif
 
