/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 2003-2004 James Bunton <james@delx.cjb.net>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * -------------------------------------------------------------------------- */


#include "msntrans.h"
#include "filetransfer.h"
#include <stdio.h>



/* Functions for all msnftp transfers */

#ifdef linux
// FIXME, implementations for other platforms (eg *BSDs) are welcome. Currently this
// will likely only work for Linux.
#include <netinet/in.h>
#include <net/if.h>             //where if_nameindex() was hiding
#include <linux/sockios.h>      //for ioctl() arg
#include <sys/ioctl.h>

//thanks linus
#define NIPQUAD(addr)   ((unsigned char *)&addr)[0], \
			((unsigned char *)&addr)[1], \
			((unsigned char *)&addr)[2], \
			((unsigned char *)&addr)[3]

void mt_msnftp_fillourip(msnftp ft)
{
	int sock = 0;
	struct ifreq ifreq;
	struct sockaddr_in *saptr = NULL;
	struct if_nameindex *iflist = NULL, *listsave = NULL;
	
#ifndef SIOCGIFADDR
	log_warn(ZONE, "Newer kernel needed for file transfer support. See msnftp.c:mt_msnftp_fillip()");
	return;
#endif
	//need a socket for ioctl()
	if( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0){
		log_warn(ZONE, "Socket error in msnftp.c:mt_msnftp_fillip()");
		return;
	}
	
	//returns pointer to dynamically allocated list of structs
	iflist = listsave = if_nameindex();
	
	//walk thru the array returned and query for each
	//interface's address
	for(iflist; *(char *)iflist != 0; iflist++){
		//copy in the interface name to look up address of
		strncpy(ifreq.ifr_name, iflist->if_name, IF_NAMESIZE);
		
		//get the address for this interface
		if(ioctl(sock, SIOCGIFADDR, &ifreq) != 0){
			log_warn(ZONE, "ioctl() error in msnftp.c:mt_msnftp_fillip()");
			return;
		}
		
		//save the addresses
		saptr = (struct sockaddr_in *)&ifreq.ifr_addr;
		
		// We don't want the loopback interface
		if(strstr(ifreq.ifr_name, "lo") != 0) continue;
		// Fill the addresses
//		if(ft->ipaddress == 0) { FIXME, we only grab the last listed IP
			ft->ipaddress = pmalloc(ft->p, 17); // Enough space to hold an IP address
			sprintf(ft->ipaddress, "%d.%d.%d.%d", NIPQUAD(saptr->sin_addr.s_addr));
			ft->port = pmalloc(ft->p, 10);
			strcpy(ft->port, "6891");
/*		}
		else if(ft->ipaddressalt == 0) {
			ft->ipaddressalt = pmalloc(ft->p, 17); // Enough space to hold an IP address
			sprintf(ft->ipaddressalt, "%d.%d.%d.%d", NIPQUAD(saptr->sin_addr.s_addr));
			ft->portalt = pmalloc(ft->p, 10);
			strcpy(ft->portalt, "11178");
			break;
		}*/
		
	}
	//free the dynamic memory kernel allocated for us
	if_freenameindex(listsave);
}

#else // linux
void mt_msnftp_fillourip(msnftp ft) {
	log_warn(ZONE, "No implementation of mt_msnftp_fillip() for your platform. Please look in msnftp.c for details.");
}
#endif // linux

void mt_msnftp_cancel(msnftp ft) {
	spool sp = spool_new(ft->p);
	
	spool_add(sp, MIME_INVITE);
	spool_add(sp, "\r\n");
	spool_add(sp, "Invitation-Command: CANCEL\r\n");
	spool_add(sp, "Invitation-Cookie: ");
	spool_add(sp, ft->invitationcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "Cancel-Code: ");
	spool_add(sp, ft->cancelcode);
	spool_add(sp, "\r\n");
	mt_msnftp_sb_sendmessage(ft, spool_print(sp));
}

char * mt_msnftp_grabfield(pool p, char *text, char *field) {
	// Looks through 'text' for a line beginning with 'field', then returns all the text after the colon
	// Expects lines formatted as below
/*
     Application-Name: File Transfer
     Application-GUID: {5D3E02AB-6190-11d3-BBBB-00C04F795683}
     Invitation-Command: INVITE
     Invitation-Cookie: 85366
     Application-File: Autoexec.bat
     Application-FileSize: 187
     Connectivity: N
*/

	char *str = 0;
	char *value = 0;
	int pos = 0;
	
	// Grab the field, and move to past the spaces & colons
	str = strstr(text, field);
	if(str == 0) {
		return 0; // Field not found
	}
	str += strlen(field);
	while(str[0] == ':' || str[0] == ' ') {
		str++;
	}
	
	// Look for the end of line
	pos = 0;
	while(str[pos] != '\r') {
		pos++;
	}
	
	// Allocate some memory and save the value
	value = pmalloc(p, pos + 1);
	strncpy(value, str, pos);
	value[pos] = 0;
	
	return value;
}

msnftp mt_msnftp_find(session s, char *invitationcookie) {
	/* Finds an invitation for a user's session, based on the cookie */
	msnftp ft = 0;
	
	if(s->ti->filetransfer == 0) return NULL;
	if(invitationcookie == 0) return NULL;
	
	ft = (msnftp) xhash_get(s->msnftps, invitationcookie);
	
	return ft;
}


void mt_msnftp_delete(msnftp ft) {
	/* Deletes a msnftp object */
	log_debug(ZONE, "Deleting msnftp object with invitationcookie: %s", ft->invitationcookie);
	if(ft->oob_m) {
		mio_close(ft->oob_m);
	}
	xhash_zap(ft->s->msnftps, ft->invitationcookie);
	pool_free(ft->p);
}



/* Functions for negotiating file transfer invitations with a private switchboard */

void mt_msnftp_sb_handle(session s, char *remoteperson, char *body, sbchat sc) {
	/* Handles invitation negotiation in the switchboard */
	char *invitationcommand = 0;
	char *invitationcookie = 0;
	msnftp ft = 0;
	
	// Check if file transfers are turned on
	if(s->ti->filetransfer == 0) return;
	
	// Parse the invitation
	invitationcommand = mt_msnftp_grabfield(s->p, body, "Invitation-Command");
	invitationcookie = mt_msnftp_grabfield(s->p, body, "Invitation-Cookie");
	log_debug(ZONE, "Parsing invitation packet: %s %s", invitationcommand, invitationcookie);
	
	if(strcmp(invitationcommand, "INVITE") == 0) {
		log_debug(ZONE, "Received INVITE for session %s", jid_full(s->id));
		// Create an invitation object to handle this
		ft = mt_msnftp_receive_create(s, remoteperson, body, sc);
		if(ft->cancelcode != 0) { // It wasn't a file transfer invite, or some other error
			mt_msnftp_cancel(ft);
			mt_msnftp_delete(ft);
			return;
		}
		ft->stage = INCOMING_INVITE_RECEIVED;
		// Send out a reply to the invitation
		log_debug(ZONE, "Sending ACCEPT for session %s", jid_full(s->id));
		mt_msnftp_receive_accept(ft);
	}
	else if(strcmp(invitationcommand, "ACCEPT") == 0) {
		ft = mt_msnftp_find(s, invitationcookie);
		if(ft == 0) return;
		switch(ft->stage) {
			case INCOMING_ACCEPT_SENT:
				mt_msnftp_receive_starttransfer(ft, body);
				break;
			
			case INCOMING_WAITING_FOR_IP:
				mt_msnftp_receive_offertoserve(ft, body);
				break;
			
			default:
				log_debug(ZONE, "Some other state");
		}
	}
	else if(strcmp(invitationcommand, "CANCEL") == 0) {
		ft = mt_msnftp_find(s, invitationcookie);
		mt_msnftp_delete(ft);
	}
}

void mt_msnftp_sb_checkprivate(msnftp ft) {
	// Guarantees a private switchboard session
	/* We need to create a new private switchboard if the switchboard we have now
		- is non-existant
		- is closed (maybe changed into a sbroom)
		- is empty (the person left)
		- has extra people in it
	*/
	sbchat sc = 0;
	session s = ft->s;
	if(	ft->privatesb == 0 ||
		ft->privatesb->state == sb_CLOSE ||
		ft->privatesb->users == NULL ||
		ft->privatesb->users->next != NULL
	  )
	{
		log_debug(ZONE, "Creating new private switchboard for invitations for session %s", jid_full(ft->s->id));
		sc = mt_chat_new(s,ft->remoteperson);
		sc->msnftp_invite_cookie = ft->invitationcookie;
		mt_stream_register(s->st,&mt_chat_xfr,(void *) sc);
		mt_cmd_xfr_sb(s->st);
		ft->privatesb = sc;
	}
}

void mt_msnftp_sb_sendmessage(msnftp ft, char *text) {
	/* Sends a switchboard message. Guarantees a private switchboard */
	mt_msnftp_sb_checkprivate(ft);
	
	if(ft->privatesb->state != sb_READY) {
		if(ft->sbbuff == 0) {
			ft->sbbuff = spool_new(ft->p);
		}
		spool_add(ft->sbbuff, text);
	}
	else {
		mt_cmd_msg(ft->privatesb->st, "N", text);
	}
}

void mt_msnftp_sb_flush(session s, char *invitationcookie) {
	// Is called when our new private switchboard is open. This sends any waiting data to that switchboard
	msnftp ft = 0;
	struct spool_node *tmp;
	
	ft = mt_msnftp_find(s, invitationcookie);
	if(ft == 0) return;
	
	// Now we need to send any buffered data
	if(ft->sbbuff == 0) return;
	ft->sbbuff->last = 0;
	while(ft->sbbuff->first != NULL) {
		tmp = ft->sbbuff->first;
		ft->sbbuff->first = tmp->next;
		mt_cmd_msg(ft->privatesb->st, "N", tmp->c);
	}
}





/* Functions for msnftp out of band streams (I could use mpstream, but it would require too much adaptation */
void mt_msnftp_oob_connect(msnftp ft) {
	// FIXME, try alternate IP address if the first one fails
	mio_connect(ft->ipaddress, j_atoi(ft->port, 6891), mt_msnftp_oob_in, (void *)ft, 0, NULL, NULL);
//	mio_connect(ft->ipaddressalt, j_atoi(ft->portalt, 11178), mt_msnftp_oob_in, (void *)ft, 0, NULL, NULL);
//	mio_connect("192.168.1.2", j_atoi(ft->port, 6891), mt_msnftp_oob_in, (void *)ft, 0, NULL, NULL);
}

void mt_msnftp_oob_listen(mio m, int state, void *arg, char *buffer, int bufsz) {
	// For listening sockets, the argument is the mti
	mti ti = (mti) arg;
	msnftp ft = (msnftp) xhash_get(ti->msnftp_listens, m->ip); // Assumes that we'll never have more than one transfer per IP
	if(ft != 0) {
		mt_msnftp_oob_in(m, state, (void *)ft, buffer, bufsz); // Forward it on, except with a sane argument
	}
}

void mt_msnftp_oob_in(mio m, int state, void *arg, char *buffer, int bufsz) {
	// Out of band handler for incoming file transfers
	msnftp ft = (msnftp) arg;
	int bufread = 0;
	
	if(ft->oob_m == 0) {
		// Connection has just been made
		log_debug(ZONE, "mt_msnftp_oob_in() - just connected");
		ft->oob_m = m;
		mt_msnftp_oob_send(ft, "VER MSNFTP\r\n");
		ft->stage = INCOMING_SENTVER;
	}
	else {
		// We already have a connection
		switch(ft->stage) {
			case INCOMING_SENTVER:
				// Check for a VER in return
				if(strstr(buffer, "VER MSNFTP") == buffer) { // At the beginning
					spool sp = spool_new(ft->p);
					spool_add(sp, "USR ");
					spool_add(sp, ft->s->user);
					spool_add(sp, " ");
					spool_add(sp, ft->authcookie);
					spool_add(sp, "\r\n");
					mt_msnftp_oob_send(ft, spool_print(sp));
					ft->stage = INCOMING_SENTUSR;
				}
				else {
					log_debug(ZONE, "Unexpected protocol error in file transfer");
				}
				break;
			
			case INCOMING_SENTUSR:
				// Check for FIL
				if(strstr(buffer, "FIL ") == buffer) { // If it's at the beginning of the string
					ft->totalbytes = j_atoi(buffer + 4, 0);
					log_debug(ZONE, "Incoming file has totalbytes: %d", ft->totalbytes);
					mt_msnftp_oob_send(ft, "TFR\r\n");
					ft->stage = INCOMING_BINARYMODE;
// 					ft->headergrabbed = -1; // FUDGE
					ft->filedata = pmalloc(ft->p, ft->totalbytes + 1); // Allocate memory for the file
				}
				else {
					log_debug(ZONE, "Unexpected protocol in file transfer");
				}
				break;
			
			case INCOMING_BINARYMODE:
				log_debug(ZONE, "Reading binary data. So far have read %d bytes of a total of %d. We have %d left in this buffer", ft->bytesreceived, ft->totalbytes, bufsz);
				while(bufread < bufsz) { // While there's still data to be read
					switch(ft->headergrabbed) {
						case 3: // The whole header, we just read the block
							while(ft->bytestonext > 0) { // While we are still in the middle of a block
								if(bufread >= bufsz) {
									break;
								}
								if(ft->bytesreceived >= ft->totalbytes) {
									break;
								}
								ft->filedata[ft->bytesreceived] = buffer[bufread];
								ft->bytesreceived++;
								bufread++;
								ft->bytestonext--;
							}
							if(ft->bytestonext == 0) {
								// Ok, new block. We have to read the header. This can be tricky because
								// it might be split across several TCP packets
								ft->headergrabbed = 0;
								log_debug(ZONE, "New header");
							}
							break;
						
						case 2: // We haven't read the last byte
							log_debug(ZONE, "Third byte");
							ft->bytestonext += (int)((unsigned char)buffer[bufread])*256;
							bufread++;
							ft->headergrabbed = 3;
							log_debug(ZONE, "New block with size %d", ft->bytestonext);
							break;
						
						case 1: // We haven't read the last two bytes
							log_debug(ZONE, "Second byte");
							ft->bytestonext = (int)((unsigned char)buffer[bufread]);
							bufread++;
							ft->headergrabbed = 2;
							break;
						
						case 0:
							bufread++;
							ft->headergrabbed = 1;
							log_debug(ZONE, "First byte");
							break;
						
// 						case -1:
// 							ft->headergrabbed = 0; // FUDGE!
// 							bufread++;
// 							break;
					}
				}
				if(ft->bytesreceived == ft->totalbytes) {
					mt_msnftp_oob_send(ft, "BYE 16777989\r\n");
					log_debug(ZONE, "File transfer finished!");
					mt_msnftp_receive_finished(ft);
					ft->stage = TRANSFERDONE;
					return;
				}
				if(ft->bytesreceived > ft->totalbytes) {
					log_error(ZONE, "File transfer received more data than was advertised. This should never happen.");
				}
				break;
			
			default:
				break;
		}
	}
}

void mt_msnftp_oob_send(msnftp ft, char *data) {
	if(ft->oob_m == NULL) {
		log_warn(ZONE, "Data lost because socket not connected yet. mt_msnftp_oob_send()");
	}
	mio_write(ft->oob_m,NULL,data,strlen(data));
}





/* Functions for being on the receiving end of a msnftp*/

msnftp mt_msnftp_receive_create(session s, char *remoteperson, char *text, sbchat sc) {
	/* Creates an msnftp object based on an incoming invitation */
	msnftp ft = 0;
	pool p = pool_new();
	
	// Check if file transfers are turned on
	if(s->ti->filetransfer == 0) return NULL;
	
	ft = pmalloc(p, sizeof(_msnftp));
	ft->s = s;
	ft->p = p;
	ft->remoteperson = pstrdup(p, remoteperson);
	ft->stage = NOTHING;
	ft->invitationcookie = mt_msnftp_grabfield(p, text, "Invitation-Cookie");
	ft->applicationguid = mt_msnftp_grabfield(p, text, "Application-GUID");
	ft->applicationfile = mt_msnftp_grabfield(p, text, "Application-File");
	ft->applicationfilesize = mt_msnftp_grabfield(p, text, "Application-FileSize");
	ft->connectivity = j_strcmp(mt_msnftp_grabfield(p, text, "Connectivity"), "N");
	
	// Create a private switchboard
	ft->privatesb = sc;
	mt_msnftp_sb_checkprivate(ft);
	
	// Null out some other variables
	ft->remoteipaddress = 0;
	ft->ipaddress = 0;
	ft->ipaddressalt = 0;
	ft->port = 0;
	ft->portalt = 0;
	ft->authcookie = 0;
	ft->sbbuff = 0;
	ft->totalbytes = 0;
	ft->bytesreceived = 0;
	ft->bytestonext = 0;
	ft->headergrabbed = 0;
	ft->filedata = 0;
	ft->oob_m = 0;
	ft->cancelcode = 0;
	
	
	if(ft->invitationcookie == 0) {
		ft->cancelcode = "FAIL";
	}
	if(strcmp(ft->applicationguid, MSNFTP_GUID) != 0) {
		ft->cancelcode = "REJECT_NOT_INSTALLED";
	}
	if(ft->applicationfile == 0) {
		ft->cancelcode = "FAIL";
	}
	if(ft->applicationfilesize == 0) {
		ft->cancelcode = "FAIL";
	}
	
	xhash_put(s->msnftps, ft->invitationcookie, ft);
	
	return ft;
}


void mt_msnftp_receive_accept(msnftp ft) {
	session s = ft->s;
	spool sp = spool_new(ft->p);
	
	if(s->ti->filetransfer == 0) return;
	
	// Construct and send the ACCEPT packet
	ft->stage = INCOMING_ACCEPT_SENT;
	spool_add(sp, MIME_INVITE);
	spool_add(sp, "Invitation-Command: ACCEPT\r\n");
	spool_add(sp, "Invitation-Cookie: ");
	spool_add(sp, ft->invitationcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "Launch-Application: FALSE\r\n");
	spool_add(sp, "Request-Data: IP-Address:\r\n");
	if(ft->connectivity == 0) {
		ft->stage = INCOMING_WAITING_FOR_IP;
	}
	spool_add(sp, "\r\n");
	mt_msnftp_sb_sendmessage(ft, spool_print(sp));
}

void mt_msnftp_receive_offertoserve(msnftp ft, char *body) {
	session s = ft->s;
	spool sp = spool_new(ft->p);
	
	if(s->ti->filetransfer == 0) return;
	
	// Grab the IP that they sent in the last invitation packet
	ft->remoteipaddress = mt_msnftp_grabfield(ft->p, body, "IP-Address");
	
	ft->authcookie = pmalloc(ft->p, 15);
	sprintf(ft->authcookie, "%d", rand()%2147483647);
	mt_msnftp_fillourip(ft);
	
	// Construct and send the ACCEPT packet
	ft->stage = INCOMING_OFFERTOSERVE_SENT;
	spool_add(sp, MIME_INVITE);
	spool_add(sp, "Invitation-Command: ACCEPT\r\n");
	spool_add(sp, "Invitation-Cookie: ");
	spool_add(sp, ft->invitationcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "Launch-Application: FALSE\r\n");
	spool_add(sp, "IP-Address: ");
	spool_add(sp, ft->ipaddress);
	spool_add(sp, "\r\n");
	spool_add(sp, "Port: ");
	spool_add(sp, ft->port);
	spool_add(sp, "\r\n");
	spool_add(sp, "Sender-Connect: TRUE\r\n");
	if(ft->ipaddressalt != 0 && ft->portalt != 0) {
		spool_add(sp, "IP-Address-Internal: ");
		spool_add(sp, ft->ipaddressalt);
		spool_add(sp, "\r\n");
		spool_add(sp, "PortX: ");
		spool_add(sp, ft->portalt);
		spool_add(sp, "\r\n");
	}
	spool_add(sp, "AuthCookie: ");
	spool_add(sp, ft->authcookie);
	spool_add(sp, "\r\n");
	spool_add(sp, "\r\n");
	
	// The next function will be mt_msnftp_oob_listen (called when they connect to our IP)
	xhash_put(s->ti->msnftp_listens, ft->remoteipaddress, ft); // Index by the ip address
	
	mt_msnftp_sb_sendmessage(ft, spool_print(sp));
}

void mt_msnftp_receive_starttransfer(msnftp ft, char *body) {
	pool p = ft->p;
	
	log_debug(ZONE, "Inside mt_msnftp_receive_starttransfer");
	
	ft->ipaddress = mt_msnftp_grabfield(p, body, "IP-Address");
	ft->ipaddressalt = mt_msnftp_grabfield(p, body, "IP-Address-Internal");
	ft->port = mt_msnftp_grabfield(p, body, "Port");
	ft->portalt = mt_msnftp_grabfield(p, body, "PortX");
	ft->authcookie = mt_msnftp_grabfield(p, body, "AuthCookie");
	
	mt_msnftp_oob_connect(ft);
	
	ft->stage = INCOMING_TRANSFERSTARTED;
}

void mt_msnftp_receive_notify(msnftp ft) {
	session s = ft->s;
	xmlnode msg;
	char *user = mt_mid2jid_full(ft->p, ft->remoteperson, s->host);
	spool body = spool_new(ft->p);
	
	spool_add(body, "You have received a file from user: ");
	spool_add(body, user);
	spool_add(body, "\n\n");
	spool_add(body, "To download this file, please go to the following address:\n");
	spool_add(body, "http://delx.cjb.net/msnfiles\n");
	
	msg = xmlnode_new_tag("message");
	xmlnode_put_attrib(msg,"to",jid_full(s->id));
	xmlnode_put_attrib(msg,"from",user);
	xmlnode_put_attrib(msg,"type","message");
	xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),spool_print(body),-1);
	
	mt_deliver(s->ti,msg);
}

void mt_msnftp_receive_save(msnftp ft) {
	FILE *out = 0;
	spool filename = spool_new(ft->p);
	int i = 0;
	
	system("mkdir -p /tmp/msnfiles");
	
	spool_add(filename, "/tmp/msnfiles/");
	spool_add(filename, ft->applicationfile);
	out = fopen(spool_print(filename), "w");
	
	for(i = 0; i < ft->totalbytes; i++) {
		fputc(ft->filedata[i], out);
	}
	fclose(out);
	out = 0;
}

void mt_msnftp_receive_finished(msnftp ft) {
	// Save the file, then notify the user that it's ready
	mt_msnftp_receive_save(ft);
	mt_msnftp_receive_notify(ft);
	mt_msnftp_delete(ft);
	ft = 0;
}
